/* This file is Copyright 2000 Level Control Systems.  See the included LICENSE.txt file for details. */

//! author="Jeremy Friesner"

#ifndef MESSAGE_TRANSCEIVER_THREAD_H
#define MESSAGE_TRANSCEIVER_THREAD_H

#include <app/Message.h>
#include <app/Messenger.h>
#include <kernel/OS.h>
#include <support/Locker.h>
#include "iogateway/AbstractMessageIOGateway.h"
#include "besupport/ThreadSafeObjectPool.h"

#define PORTABLE_MESSAGES_RECEIVED 'PmRR'

//:Type representing any function that will return a newly allocated AbstractMessageIOGateway object that uses the given PortableDataIO object.
typedef AbstractMessageIOGateway* (*GatewayFactoryFunc)(PortableDataIO * io);

//:This is a utility class to help interface multithreaded BeOS
//:code more easily to a TCP socket that is sending and receiving
//:PortableMessages.   Only compiles under BeOS!
// It helps by allowing you to send and receive
// PortableMessages without having to worry about calling select(),
// or any blocking issues.
//
// Usage:  1. Create a new MessageTransceiverThread object.
//         2. Call StartThread(), StartAcceptThread(), StartServerThread(), or StartConnectThread().
//         3. Whenever you wish to send a PortableMessage over the TCP socket,
//            simply call SendOutgoingMessage().  
//         4. Whenever a PortableMessage is received, a PORTABLE_MESSAGES_RECEIVED
//            BMessage will be sent to the BMessenger you specified in StartThread().
//            When you get this, you can respond by calling GetNextIncomingMessage()
//            on this object to retrieve the PortableMessage.
//         5. When you are done, just delete this object normally.  If you wish,
//            you can call WaitForAllMessagesToBeSent() first, in order to ensure
//            that all pending outbound messages get delivered.
//         6. That's it!
//
class MessageTransceiverThread
{
public:
   // Creates a receiver thread without a reply target.  You should
   // call SetReplyTarget() before calling StartThread()...
   MessageTransceiverThread(int32 threadPri = B_NORMAL_PRIORITY);

   // (replyTarget) specified where you want the PORTABLE_MESSAGES_RECEIVED (and other messages) sent to .
   // (threadPri) is the priority the TCP thread should run at.
   MessageTransceiverThread(const BMessenger & replyTarget, int32 threadPri = B_NORMAL_PRIORITY);

   // May be called at any time
   virtual ~MessageTransceiverThread();

   // Creates a TCP send/receive thread that connects to the given hostname and port.
   // (hostname) is the Internet hostname of the computer to connect to.
   // (port) is the port number to connect on.
   // (optConnectMessage), if specified, will be sent via your BMessenger when the connection 
   //                      is complete.  (A copy of this message is made).
   // (optDisconnectMessage), if specified, will be sent via your BMessenger when the connection 
   //                         fails or is broken. (A copy of this message is made).
   status_t StartConnectThread(const char * hostname, int port, 
                               const BMessage * optConnectMessage = NULL, const BMessage * optDisconnectMessage = NULL);

   // Creates a TCP send/receive thread that waits for connections on the given port.
   // If (port) is 0, then the system will choose a port to accept on.  The port number it
   // chooses can be discovered by calling GetPort().
   // (acceptMultiple), if true, will allow subsequent connections to be accepted after the first one is closed.
   //                   In no case will there be more than one connection active at a time, however.
   // (optConnectMessage), if specified, will be sent via your BMessenger when the connection 
   //                      is complete.  (A copy of this message is made).
   // (optDisconnectMessage), if specified, will be sent via your BMessenger when the connection 
   //                         fails or is broken. (A copy of this message is made).
   status_t StartAcceptThread(int port, bool acceptMultiple,
                              const BMessage * optConnectMessage = NULL, const BMessage * optDisconnectMessage = NULL);

   // Creates a TCP accept thread that waits for connections on the given port.
   // If (port) is 0, then the system will choose a port to accept on.  The port number it
   // chooses can be discovered by calling GetPort().
   // The difference between calling this call and StartAcceptThread() lies in how the
   // incoming connection is handled.  With StartAcceptThread(), the incoming connection
   // becomes the 'current connection' for this object, and PortableMessages are sent and
   // received as usual.  With this method, on the other hand, an incoming connection is
   // accepted and placed into an incoming-connections queue, and the (connectMessage) you
   // specify is sent to your handler.  Your handler should then call GetNextIncomingConnection() 
   // on this object to retrieve the socket for the new connection, which can be handled with
   // a new MessageTransceiverThread or however you like. 
   // If (acceptMultiple) is false, then only the first incoming connection will ever be accepted.
   status_t StartServerThread(int port, const BMessage & connectMessage, bool acceptMultiple=true);

   // Creates and activates the TCP send/receive background thread.
   // (gateway) is the AbstractMessageIOGateway the thread should use to send and receive data.
   //           In general, a AbstractMessageIOGateway object with a non-blocking TCPSocketDataIO
   //           that uses your socket is the way to go here.
   // (optConnectMessage), if specified, will be sent via your BMessenger when the connection 
   //                      is complete.  (A copy of this message is made).
   // (optDisconnectMessage), if specified, will be sent via your BMessenger when the connection 
   //                         fails or is broken. (A copy of this message is made).
   status_t StartThread(AbstractMessageIOGateway * gateway, int datafd, const BMessage * optDisconnectMessage = NULL);
   
   // If there are any incoming messages pending, places a reference to the next one into (setRef) and returns B_NO_ERROR.
   // Otherwise returns B_ERROR.
   status_t GetNextIncomingMessage(PortableMessageRef & setRef);

   // Enqueues (msgRef) for immediate output over the TCP connection.
   // Returns B_NO_ERROR on success, B_ERROR if no more messages can be sent (e.g. connection is broken)
   // Note that it is NOT safe to keep your own local copies of (msgRef) around
   // after this method returns (because PortableMessageRef isn't thread safe)
   // NOTE:  If you want to use object pools with your (msgRef), make sure that the
   //        pools you use are ThreadSafeObjectPools objects, or you will suffer from
   //        nasty race conditions.  The GetThreadSafeMessagePool() and GetThreadSafeMessageRefPool()
   //        methods in this class return appropriate pool objects to use here.
   status_t SendOutgoingMessage(const PortableMessageRef & msgRef);

   // Waits until all messages that were on the queue to be sent are sent, and then returns B_NO_ERROR.
   // If for some reason it could not wait, B_ERROR is returned.
   void WaitForAllMessagesToBeSent();

   // Returns the port number that the accept thread is accepting messages on. 
   // This number is only meaningful if StartAcceptThread() was called.
   int GetPort() const;

   // Sets or resets the reply target for this thread
   void SetReplyTarget(const BMessenger & target);

   // If you wish to use a type of AbstractMessageIOGateway other than the default type,
   // you can set a callback to your own factory function here.  The factory function should
   // return a new AbstractMessageIOGateway object (or subclass thereof), or NULL on failure.
   // The returned object becomes property of this MessageTransceiverThread object.
   void SetGatewayFactoryFunc(GatewayFactoryFunc func);

   // Call this if you wish to receive a BMessage every time the outgoing message queue empties.  
   // (msg) should point to the message to send, or be NULL to turn off empty-queue-messaging.
   // A copy of (*msg) will be made.
   void SetOutgoingQueueDrainedMessage(const BMessage * msg);

   // This call is only useful for MessageTransceiverThreads that are running in server mode
   // (i.e. that you have called StartServerThread() on).  It will retrieve the next queued
   // incoming TCP connection socket and copy it into (returnSocketFD).  If a socket is 
   // returned to you, your code becomes the owner of the socket.
   // If there is no incoming connection ready, this method will return B_ERROR.
   status_t GetNextIncomingConnection(int & returnSocketFD);

   // Returns a pointer to a singleton thread-safe message pool.
   // Equivalent to ::GetMessagePool(), but for use in multithreaded BeOS environments.
   static ThreadSafeObjectPool<PortableMessage> * GetMessagePool() {return &_msgPool;}

   // Returns a pointer to a singleton thread-safe message ref pool.
   // Equivalent to ::GetMessageRefPool(), but for use in multithreaded BeOS environments.
   static ThreadSafeObjectPool<PortableMessageRef::RefCountMem> * GetMessageRefPool() {return &_msgRefPool;}

private:
   AbstractMessageIOGateway * CreateIOGateway(int socketfd) const;

   status_t StartThread(thread_func func, const char * threadName);
   void NotifyThread();
   status_t SetSocketBlockingEnabled(int socket, bool blocking);
   status_t CreateNotificationMechanism();
   static int32 TCPThreadStub(void * d) {return ((MessageTransceiverThread *)d)->TCPThreadLoop();}
   static int32 ConnectThreadStub(void * d) {return ((MessageTransceiverThread *)d)->ConnectThreadLoop();}
   void AcquireMessagesPendingLock();
   void ReleaseMessagesPendingLock();
   int32 TCPThreadLoop();
   int32 ConnectThreadLoop();

   volatile thread_id _tcpThread;
   BMessenger _replyTarget;
   BMessage _notifyMsg;
   BLocker _lock;   

   // _gateway and _datafd are linked.  If _gateway if non-NULL, _datafd must be a valid fd.
   // And if _gateway is NULL, _datafd must be < 0.
   AbstractMessageIOGateway * _gateway;
   int _datafd;    // the fd we actually transfer messages over

   int _acceptfd;  // the fd we wait for connections on (or -1 if we aren't doing that)

   int _notifyfd;  // the fd we use to wake up our thread (hacksville!)
   int _waitfd;    // the fd the thread uses to be woken up (hack hack!)

   PortableString _connectTo;  // used by StartConnectThread()
   int _port;
   int32 _threadPri;
 
   bool _isConnecting;
   bool _isServer;      // true iff StartServerThread() was called

   BMessage _onConnect;
   bool _sendOnConnect;

   BMessage _onDisconnect;
   bool _sendOnDisconnect;

   BMessage _onEmpty;
   bool _sendOnEmpty;

   bool _acceptMultiple;

   sem_id _messagesPendingSemaphore;
   int32 _messagesPendingBenaphore;

   GatewayFactoryFunc _userGatewayFunc;

   PortableQueue<int> _incomingConnections;  // used when in server mode

   static ThreadSafeObjectPool<PortableMessage> _msgPool;
   static ThreadSafeObjectPool<PortableMessageRef::RefCountMem> _msgRefPool;
};

#endif
