/* This file is Copyright 2000 Level Control Systems.  See the included LICENSE.txt file for details. */

//! author="Jeremy Friesner"

#ifndef PORTABLE_MESSAGE_IO_GATEWAY
#define PORTABLE_MESSAGE_IO_GATEWAY

#include "dataio/PortableDataIO.h"
#include "message/PortableMessage.h"
#include "refcount/PortableRefCount.h"
#include "iogateway/AbstractMessageIOGateway.h"

// Encoding IDs.  Only ENCODING_DEFAULT is supported for now!
enum {
   ENCODING_DEFAULT = 1164862256, // 'Enc0',  // just plain ol' flattened PortableMessage objects
   ENCODING_ZIP,                              // compressed with ZIP compression:  not supported (yet)
   ENCODING_BLOWFISH                          // encrypted with BLOWFISH:  not supported (yet)  (maybe never)
};

//:A "gateway" object that knows how to send/receive PortableMessages over a wire
//:(via a provided PortableDataIO object). 
// May be subclassed to change the byte-level
// protocol, or used as is if the default protocol is desired.
//
// The default protocol format used by this class is:
//    1. 4 bytes (uint32) indicating the flattened size of the message
//    2. 4 bytes (uint32) indicating the encoding type (should always be ENCODING_DEFAULT for now)
//    3. n bytes of flattened PortableMessage (where n is the value specified in 1)
//    4. goto 1 ...
//
class PortableMessageIOGateway : public AbstractMessageIOGateway
{
public:
   // This object becomes the owner of (io) and will delete it later.
   PortableMessageIOGateway(PortableDataIO * io, long encoding = ENCODING_DEFAULT);
   virtual ~PortableMessageIOGateway();

   // Writes some of our outgoing messages to the wire.
   // Not guaranteed to write all outgoing messages (it will try not to block)
   // Returns B_NO_ERROR usually, or B_ERROR if the connection has been broken
   // or some other catastrophic condition has occurred.
   virtual status_t DoOutput();

   // Reads some more bytes from the wire.  
   // If enough bytes have been read to assemble one or more new PortableMessages,
   // the new PortableMessages will be available after this method returns by
   // calling GetNextIncomingMessage().  
   // Tries not to block, but may (depending on implementation)
   // Returns B_NO_ERROR usually, or B_ERROR if the connection has been broken
   // or some other catastrophic condition has occurred.
   virtual status_t DoInput();

   // Returns true iff there are any PortableMessages that have been
   // added (with AddOutgoingMessage()) but not already sent across the PortableDataIO.
   // Always returns false if the connection has been closed or hosed.
   virtual bool HasBytesToOutput() const;

protected:
   // Should return the total size of the body buffer for sending this (msg).
   // Default implementation returns msg.FlattenedSize().
   virtual size_t FlattenedMessageBodySize(const PortableMessage & msg) const;

   // Flattens (msg) into (buffer) and (header).
   // (buffer) is FlattenedMessageSize() bytes long.  This should have header info written to it (including body size!)
   // (header) is HeaderSize() bytes long.  This should have the body data written into it
   // Returns B_NO_ERROR if the message should be sent, B_ERROR if this message should 
   // be dropped.
   // Default implementation always returns true.
   virtual status_t FlattenMessage(const PortableMessage & msg, uint8 * header, uint8 * buffer) const;

   // Unflattens (msg) from (buffer) and (header).
   // (header) is HeaderSize() bytes long.
   // (buffer) is (bufSize) bytes long.
   virtual status_t UnflattenMessage(PortableMessage & setMsg, const uint8 * header, const uint8 * buf, size_t bufSize) const;
 
   // Returns the size of the pre-flattened-message header section, in bytes.
   // The default format has an 8-byte header (4 bytes for encoding ID, 4 bytes for message size)
   virtual size_t HeaderSize() const;

   // Extracts and returns (in setSize) the dynamic buffer size from the given header.
   // Note that the returned size should NOT include the header bytes!
   // (headerBuf) is HeaderSize() bytes long.
   // Returns B_NO_ERROR iff successful.         
   virtual status_t GetBodySize(const uint8 * header, size_t & setSize) const;

   // This method is called after GetBodySize() returns an error code.  (i.e.
   // when the header bytes we read in were not valid)
   // It should return B_NO_ERROR if the gateway should try to recover from
   // the invalid header, or B_ERROR if it should just give up and close the connection.
   // If it returns B_NO_ERROR, it may set the (setFirstValidByte) parameter to the
   // index of the first byte in the header buffer to start reading the next header
   // from, or leave (setFirstValidByte) unchanged to throw out all the bytes in the 
   // current header buffer and start afresh.
   // Default implementation just returns B_ERROR.
   // (headerBuf) the header bytes that were read in.
   // (setFirstValidByte) the length of the header.  May be set as described above.
   virtual status_t RecoverFromInvalidHeader(const uint8 * headerBuf, size_t & setFirstValidByte) const;
   
   // This method is called after UnflattenMessage() returns an error code (i.e.
   // when the body bytes we read in were not valid)
   // It should return B_NO_ERROR if the gateway should try to recover from the
   // invalid message, or B_ERROR if it should just give up and close the connection.
   // If B_NO_ERROR is returned, the the gateway will begin trying to read the
   // next header from the bytes immediately after these.
   virtual status_t RecoverFromInvalidMessage(const uint8 * messageBuf, size_t messageBufSize) const;

private:
   uint8 * _sendBuffer;
   size_t  _sendBufferSize;
   int32   _sendBufferOffset;
   size_t  _currentSendMessageSize;  // includes header bytes

   uint8 * _recvBuffer;
   size_t  _recvBufferSize;
   int32   _recvBufferOffset;
   size_t  _currentRecvMessageSize;  // includes header bytes

   int     _encoding;
};

#endif
