/* This file is Copyright 2000 Level Control Systems.  See the included LICENSE.txt file for details. */

//! author="Jeremy Friesner"

/******************************************************************************
/
/   File:       PortableMessage.h
/
/   Description:   OS-independent version of Be's BMessage class.
/         Doesn't contain all of BMessage's functionality,
/         but keeps enough of it to be useful as a portable
/         data object (dictionary) in cross-platform apps.
/
*******************************************************************************/

#ifndef _PORTABLEMESSAGE_H
#define _PORTABLEMESSAGE_H

#include "refcount/PortableRefCount.h"
#include "support/PortableTypeDefs.h"
#include "support/PortableFlattenable.h"
#include "support/PortablePoint.h"
#include "support/PortableRect.h"
#include "string/PortableString.h"
#include "hashtable/PortableHashtable.h"
#include "refcount/PortableRefCount.h"

class PortableMessage;
typedef PortableRef<PortableMessage> PortableMessageRef;

/*--------------------------------------------------------------------*/
/* --------- PortableMessage class----------------------------------- */

class PortableMessage;

// This function returns a pointer to a singleton ObjectPool that can be 
// used to minimize the number of PortableMessage allocations and deletions by
// recycling the PortableMessage objects.  Note that the returned ObjectPool is
// NOT THREAD SAFE, so you probably shouldn't use it in multithreaded
// programs.
PortableMessageRef::ItemPool * GetMessagePool();
PortableMessageRef::RefPool * GetMessageRefPool();

// Convenience method:  Gets a message from the message pool and returns it.
PortableMessage * GetMessageFromPool(uint32 what = 0L);

// Convenience method:  Gets a message from the message pool, makes it equal to (copyMe), and returns it. 
PortableMessage * GetMessageFromPool(const PortableMessage & copyMe);

// these declarations are for internal use only
class AbstractDataArray;
typedef PortableHashtableIterator<PortableString, AbstractDataArray *> _eitr;   

//: This is an iterator that allows you to efficiently iterate over the field
//: names in a PortableMessage. 
// It can be allocated solely by the PortableMessage::GetFieldNameIterator() method.
class PortableMessageFieldNameIterator
{
public:
   // Creates an "empty" iterator
   PortableMessageFieldNameIterator();

   // Creates an iterator identical to (copyMe)
   PortableMessageFieldNameIterator(const PortableMessageFieldNameIterator & copyMe);

   // Sets this iterator to be the same as (rhs)
   PortableMessageFieldNameIterator & operator = (const PortableMessageFieldNameIterator & rhs);

   // Deletes the iterator
   ~PortableMessageFieldNameIterator();

   // Returns true iff there are more field names available in the list.
   bool HasMoreFieldNames() const;

   // Places the next field name into (name) and bumps the iterator.
   // Returns true iff successful, false if no more names are left.
   bool GetNextFieldName(PortableString & name);

   // As above, only more efficient as no data is copied.  (Note that
   // the returned string may become invalid when the hashtable is changed, though!)
   const char * GetNextFieldName();

   // Places the next field name into (name) and returns true.
   // If there are no more field names left, returns false.
   bool PeekNextFieldName(PortableString & name) const;

   // As above, only returns a pointer instead of an object (see note above)
   const char * PeekNextFieldName() const;

private:
   friend class PortableMessage;

   PortableMessageFieldNameIterator(const PortableMessage & owner, _eitr * iter, type_code tc)
    : _owner(&owner), _iter(iter), _typeCode(tc) {NoteAllocated(this, "PortableMessageFieldNameIterator");}

   const PortableMessage * _owner;
   _eitr * _iter;
   type_code _typeCode;
};

// Version number of the PortableMessage serialization protocol.
// Will be the first four bytes of each serialized PortableMessage buffer.
// Only buffers beginning with version numbers between these two 
// constants (inclusive) will be Unflattened as PortableMessages.
#define OLDEST_SUPPORTED_PROTOCOL_VERSION 1347235888 // 'PM00'
#define CURRENT_PROTOCOL_VERSION          1347235888 // 'PM00'

//: This class is similar to a BMessage, but portable.  
// It only acts as a serializable data container; it does not have
// include any threading capabilities.  Unless otherwise noted, all
// methods behave similarly to their BMessage counterparts.  (Exception:
// the only error codes returned are B_ERROR and B_NO_ERROR)
class PortableMessage : public PortableFlattenable 
{
public:
   uint32   what;

   PortableMessage();
   PortableMessage(uint32 what);
   PortableMessage(const PortableMessage &a_message);
   virtual ~PortableMessage();

   PortableMessage &operator=(const PortableMessage &msg);

   status_t GetInfo(type_code typeRequested, int32 which, PortableString & setName, type_code *typeReturned, int32 *count = NULL, bool * fixed_size = NULL) const;
   status_t GetInfo(const char *name, type_code *type, int32 *c) const;
   status_t GetInfo(const char *name, type_code *type, bool *fixed_size) const;
   status_t GetInfo(const char *name, type_code *type, int32 *c, bool *fixed_size) const;

   int32 CountNames(type_code type = B_ANY_TYPE) const;
   bool IsEmpty() const;
   void PrintToStream() const;

   status_t Rename(const char *old_entry, const char *new_entry);

   /* Flattening/Unflattening data */
   virtual bool IsFixedSize() const {return false;}
   virtual type_code TypeCode() const {return B_MESSAGE_TYPE;}
   virtual size_t FlattenedSize() const;
   virtual void Flatten(uint8 *buffer) const;
   virtual status_t Unflatten(const uint8 *buf, size_t size);

   status_t AddString(const char *name, const char *a_string);
   status_t AddString(const char *name, const PortableString & a_string);
   status_t AddInt8(const char *name, int8 val);
   status_t AddInt16(const char *name, int16 val);
   status_t AddInt32(const char *name, int32 val);
   status_t AddInt64(const char *name, int64 val);
   status_t AddBool(const char *name, bool a_boolean);
   status_t AddFloat(const char *name, float a_float);
   status_t AddDouble(const char *name, double a_double);
   status_t AddMessage(const char *name, const PortableMessage & msg);
   status_t AddMessage(const char *name, PortableMessageRef msgRef); 
   status_t AddPointer(const char *name, const void * ptr);
   status_t AddPoint(const char *name, const PortablePoint & point);
   status_t AddRect(const char *name, const PortableRect & rect);
   status_t AddFlat(const char *name, const PortableFlattenable &obj);

   //:Note:  If you are adding B_MESSAGE_TYPE, then (data) must point to a PortableMessageRef
   //     object, and NOT a PortableMessage object, or flattened-PortableMessage-buffer!
   //     (This is inconsistent with BMessage::AddData, which expects a flattened BMessage buffer.  Sorry!)
   // As of MUSCLE v1.41, (data) may be NULL, in which case default objects (or uninitialized space) is added.
   status_t AddData(const char *name, type_code type, const void *data, size_t numBytes);

   // These methods are like Add*(), but add the new item at the beginning
   // of the array instead of the end.  (these methods not found in BMessage)
   status_t PrependString(const char *name, const char *a_string);
   status_t PrependString(const char *name, const PortableString & a_string);
   status_t PrependInt8(const char *name, int8 val);
   status_t PrependInt16(const char *name, int16 val);
   status_t PrependInt32(const char *name, int32 val);
   status_t PrependInt64(const char *name, int64 val);
   status_t PrependBool(const char *name, bool a_boolean);
   status_t PrependFloat(const char *name, float a_float);
   status_t PrependDouble(const char *name, double a_double);
   status_t PrependMessage(const char *name, const PortableMessage & msg);
   status_t PrependMessage(const char *name, PortableMessageRef msgRef); 
   status_t PrependPointer(const char *name, const void * ptr);
   status_t PrependPoint(const char *name, const PortablePoint & point);
   status_t PrependRect(const char *name, const PortableRect & rect);
   status_t PrependFlat(const char *name, const PortableFlattenable &obj);
   status_t PrependData(const char *name, type_code type, const void *data, size_t numBytes);

   status_t RemoveData(const char *name, int32 index = 0);
   status_t RemoveName(const char *name);
   void MakeEmpty();

   status_t FindString(const char *name, int32 index, const char ** setMe) const;
   status_t FindString(const char *name, const char ** setMe) const {return FindString(name, 0, setMe);}
   status_t FindString(const char *name, PortableString & setMe) const {return FindString(name, 0, setMe);}
   status_t FindString(const char *name, int32 index, PortableString & setMe) const;
   status_t FindInt8(const char *name, int8 *value) const {return FindInt8(name, 0, value);}
   status_t FindInt8(const char *name, int32 index, int8 *val) const;
   status_t FindInt16(const char *name, int16 *value) const {return FindInt16(name, 0, value);}
   status_t FindInt16(const char *name, int32 index, int16 *val) const;
   status_t FindInt32(const char *name, int32 *value) const {return FindInt32(name, 0, value);}
   status_t FindInt32(const char *name, int32 index, int32 *val) const;
   status_t FindInt64(const char *name, int64 *value) const {return FindInt64(name, 0, value);}
   status_t FindInt64(const char *name, int32 index, int64 *val) const;
   status_t FindBool(const char *name, bool *value) const {return FindBool(name, 0, value);}
   status_t FindBool(const char *name, int32 index, bool *value) const;
   status_t FindFloat(const char *name, float *f) const {return FindFloat(name, 0, f);}
   status_t FindFloat(const char *name, int32 index, float *f) const;
   status_t FindDouble(const char *name, double *d) const {return FindDouble(name, 0, d);}
   status_t FindDouble(const char *name, int32 index, double *d) const;
   status_t FindMessage(const char *name, PortableMessage &msg) const {return FindMessage(name, 0, msg);}
   status_t FindMessage(const char *name, int32 index, PortableMessage &msg) const;
   status_t FindMessage(const char *name, PortableMessageRef &msgRef) const {return FindMessage(name, 0, msgRef);}
   status_t FindMessage(const char *name, int32 index, PortableMessageRef &msgRef) const;
   status_t FindPointer(const char *name, void ** ptr) const {return FindPointer(name, 0, ptr);}
   status_t FindPointer(const char *name, int32 index, void ** ptr) const;
   status_t FindPoint(const char *name, PortablePoint & point) const {return FindPoint(name, 0, point);}
   status_t FindPoint(const char *name, int32 index, PortablePoint & point) const;
   status_t FindRect(const char *name, PortableRect & rect) const {return FindRect(name, 0, rect);}
   status_t FindRect(const char *name, int32 index, PortableRect & point) const;
   status_t FindFlat(const char *name, PortableFlattenable &obj) const {return FindFlat(name, 0, obj);}
   status_t FindFlat(const char *name, int32 index, PortableFlattenable &obj) const;

   //:Note:  If you are retrieving B_MESSAGE_TYPE, then (data) will be set to point to a PortableMessageRef
   //     object, and NOT a PortableMessage object, or flattened-PortableMessage-buffer!
   //     (This is inconsistent with BMessage::FindData, which returns a flattened BMessage buffer.  Sorry!)
   status_t FindData(const char *name, type_code type, const void **data, size_t *numBytes) const {return FindData(name, type, 0, data, numBytes);}
   status_t FindData(const char *name, type_code type, int32 index, const void **data, size_t *numBytes) const;

   //:As above, only this returns a pointer to an array that you can write directly into, for efficiency.  
   //You should be very careful with this method, though, as you will be writing into the guts of this PortableMessage.
   //The returned pointer is not guaranteed to be valid after the PortableMessage is modified by any method!
   status_t FindDataPointer(const char *name, type_code type, void **data, size_t *numBytes) const {return FindDataPointer(name, type, 0, data, numBytes);}
   status_t FindDataPointer(const char *name, type_code type, int32 index, void **data, size_t *numBytes) const;

/* Replacing data */
   status_t ReplaceString(const char *name, const PortableString & newString) {return ReplaceString(name, 0, newString);}
   status_t ReplaceString(const char *name, int32 index, const PortableString & newString);
   status_t ReplaceString(const char *name, const char * newString) {return ReplaceString(name, 0, newString);}
   status_t ReplaceString(const char *name, int32 index, const char * newString) {return ReplaceString(name, index, PortableString(newString));}

   status_t ReplaceInt8(const char *name, int8 val) {return ReplaceInt8(name, 0, val);}
   status_t ReplaceInt8(const char *name, int32 index, int8 val);
   status_t ReplaceInt16(const char *name, int16 val) {return ReplaceInt16(name, 0, val);}
   status_t ReplaceInt16(const char *name, int32 index, int16 val);
   status_t ReplaceInt32(const char *name, int32 val) {return ReplaceInt32(name, 0, val);}
   status_t ReplaceInt32(const char *name, int32 index, int32 val);
   status_t ReplaceInt64(const char *name, int64 val) {return ReplaceInt64(name, 0, val);}
   status_t ReplaceInt64(const char *name, int32 index, int64 val);
   status_t ReplaceBool(const char *name, bool a_bool) {return ReplaceBool(name, 0, a_bool);}
   status_t ReplaceBool(const char *name, int32 index, bool a_bool);
   status_t ReplaceFloat(const char *name, float a_float) {return ReplaceFloat(name, 0, a_float);}
   status_t ReplaceFloat(const char *name, int32 index, float a_float);
   status_t ReplaceDouble(const char *name, double a_double) {return ReplaceDouble(name, 0, a_double);}
   status_t ReplaceDouble(const char *name, int32 index, double a_double);
   status_t ReplacePointer(const char *name, const void * ptr) {return ReplacePointer(name, 0, ptr);}
   status_t ReplacePointer(const char *name, int32 index, const void * ptr);
   status_t ReplacePoint(const char *name, const PortablePoint & point) {return ReplacePoint(name, 0, point);}
   status_t ReplacePoint(const char *name, int32 index, const PortablePoint & point);
   status_t ReplaceRect(const char *name, const PortableRect & rect) {return ReplaceRect(name, 0, rect);}
   status_t ReplaceRect(const char *name, int32 index, const PortableRect & rect);
   status_t ReplaceMessage(const char *name, const PortableMessage &msg) {return ReplaceMessage(name, 0, msg);}
   status_t ReplaceMessage(const char *name, int32 index, const PortableMessage &msg);
   status_t ReplaceMessage(const char *name, PortableMessageRef msgRef) {return ReplaceMessage(name, 0, msgRef);}
   status_t ReplaceMessage(const char *name, int32 index, PortableMessageRef msgRef);
   status_t ReplaceFlat(const char *name, const PortableFlattenable &obj) {return ReplaceFlat(name, 0, obj);}
   status_t ReplaceFlat(const char *name, int32 index, const PortableFlattenable &obj);

   //:Note:  If you are replacing B_MESSAGE_TYPE, then (data) must point to a PortableMessageRef
   //     object, and NOT a PortableMessage object, or flattened-PortableMessage-buffer!
   //     (This is inconsistent with BMessage::ReplaceData, which expects a flattened BMessage buffer.  Sorry!)
   status_t ReplaceData(const char *name, type_code type, const void *data, size_t data_size) {return ReplaceData(name, type, 0, data, data_size);}
   status_t ReplaceData(const char *name, type_code type, int32 index, const void *data, size_t data_size);

   bool HasName(const char * fieldName, type_code type = B_ANY_TYPE) const;

   // Take the data under (name) in this message, and moves it into (moveTo). 
   // Any data that was under (name) in (moveTo) will be replaced.
   status_t MoveName(const char * name, PortableMessage &moveTo);

   // Take the data under (name) in this message, and copies it into (moveTo). 
   // Any data that was under (name) in (moveTo) will be replaced.
   status_t CopyName(const char * name, PortableMessage &copyTo) const;

   // Returns an iterator that iterates over the names of the fields in this
   // message.  If (type) is B_ANY_TYPE, then all field names will be included
   // in the traversal.  Otherwise, only names of the given type will be included.
   PortableMessageFieldNameIterator GetFieldNameIterator(type_code type = B_ANY_TYPE) const;

   // This method sets the pools that the PortableMessage class uses by default in situations where
   // now pools are available (e.g. from a passed-in PortableMessageRef or somesuch).  By default,
   // no pooling is done in these situations.  For a minor CPU-efficiency boost in single-threaded
   // programs, you can call SetInternalDefaultPools(GetMessagePool(), GetMessageRefPool()).  Don't
   // do this for multi-threaded programs, though, as it will cause race conditions!  
   // (For multi-threaded Be programs, you can use MessageTransceiverThread::GetMessagePool() and
   // MessageTransceiverThread::GetMessageRefPool() instead, as these will do appropriate locking
   // to guarantee thread-safety)
   static void SetInternalDefaultPools(PortableMessageRef::ItemPool * messagePool, PortableMessageRef::RefPool *messageRefPool);

private:
   // Helper functions!

   status_t AddData(const char *name, type_code type, const void *data, size_t numBytes, bool prepend);

   // Given a known type_code, returns the size of an item of that type.
   // Returns zero if items of the given type are variable length.
   status_t GetElementSize(type_code type, size_t & setSize) const;

   AbstractDataArray * GetArray(const char * arrayName, type_code etc) const {return GetArray(PortableString(arrayName), etc);}
   AbstractDataArray * GetArray(const PortableString & arrayName, type_code etc) const;

   AbstractDataArray * GetOrCreateArray(const char * arrayName, type_code tc) {return GetOrCreateArray(PortableString(arrayName), tc);}
   AbstractDataArray * GetOrCreateArray(const PortableString & arrayName, type_code tc);

   status_t AddFlatAux(const char * name, const PortableFlattenable & flat, type_code etc, bool prepend);
   status_t AddDataAux(const char * name, const uint8 * data, size_t size, type_code etc, bool prepend);

   status_t FindFlatAux(const char * name, int32 index, PortableFlattenable & flat, type_code etc) const;
   status_t FindDataAux(const char * name, int32 index, const uint8 ** setData, size_t * setSize, type_code etc) const;
   status_t FindDataItemAux(const char * name, int32 index, type_code tc, void * setValue, size_t valueSize) const;

   status_t ReplaceFlatItemAux(const char * name, int32 index, const PortableFlattenable & flat, type_code tc);
   status_t ReplaceDataAux(const char * name, int32 index, uint8 * dataBuf, size_t bufSize, type_code tc);
 
   // Iterator support methods
   friend class PortableMessageFieldNameIterator;
   const char * GetNextFieldName(_eitr * iter, type_code tc) const;
   const char * PeekNextFieldName(_eitr * iter, type_code tc) const;
   PortableHashtable<PortableString, AbstractDataArray *> _entries;   
};

#endif /* _PORTABLEMESSAGE_H */

