/* This file is Copyright 2000 Level Control Systems.  See the included LICENSE.txt file for details. */

//! author="Jeremy Friesner"

#ifndef PORTABLE_REFCOUNT_H
#define PORTABLE_REFCOUNT_H

#include <new.h>
#include "support/PortableTypeDefs.h"
#include "support/PortableMacros.h"
#include "pool/ObjectPool.h"

// Multithread safety is provided by two macros, ATOMIC_PREINCREMENT and ATOMIC_PREDECREMENT.
// For PortableRefs to work safely in multithreaded environments, these macros
// must be defined to increment or decrement their argument ATOMICALLY and return
// the resulting value.  Note that these macros are defined correctly for BeOS
// only... for other systems they are defined as the regular (non-atomic) match
// operators.  So if you are porting to another OS and are using multithreaded
// code, you will need to #define these to the appropriate functions of your OS.
#ifdef __BEOS__
# define ATOMIC_PREINCREMENT(c) (atomic_add(&c,1)+1)
# define ATOMIC_PREDECREMENT(c) (atomic_add(&c,-1)-1)
#else
# define ATOMIC_PREINCREMENT(c) (++c)  // not thread safe!!!
# define ATOMIC_PREDECREMENT(c) (--c)  // not thread safe!!!
#endif

//:This class implements reference counting on any object or array
//:you give it, so that your object will be automatically deleted when 
//:your last reference to it goes away.
template <class Item> class PortableRef
{
public:
   class RefCountMem;
   typedef ObjectPool<RefCountMem> RefPool;  // type of an ObjectPool of our internal data structures
   typedef ObjectPool<Item> ItemPool;        // type of an ObjectPool of user data structures

   // Creates a new reference-count for the given item.
   // Once this reference count has been created, (item) will be automatically
   // deleted when the last PortableRef that references it has dissappeared.
   // Note that it is an error to call this constructor twice with the same (item).
   // To obtain additional tracked references to (item), use PortableRef's copy 
   // constructor, or the PortableRef equals operator.
   //
   // (refPool) may be NULL, or if you wish to be more efficient, it may point to a RefPool
   // that you are using to recycle ref-count memory for this PortableRef class.  (Note that
   // the calling code retains ownership of the RefPool object)
   //
   // (itemPool) may be NULL, or if you wish to be more efficient, it may point to a ItemPool
   // that you are using to recycle your (item)s.  (Note that
   // the calling code retains ownership of the ItemPool object)
   //
   // If (item) is a pointer to a single item, and not an array of items, then
   // the default value of one is appropriate.
   //
   // (numItems) is the number of (Item) objects that (item) points to.  In general,
   // this will be one, but for arrays it might be more.
   //
   // If (doRefCounting) is set false, then (item) will NOT be deleted automatically.
   // It then remains the responsibility of the calling code to delete it.  (This is useful
   // if e.g. you need to put a non-new'd item into a PortableRef, and want to 
   // handle the cleanup issues yourself)
   //
   // Usage example: 
   //
   // PortableRef<Foo> ref(new Foo());
   // for (int i=0; i<10; i++) _vectors[i].AddItem(ref);
   // (the Foo object will be deleted automagically at the right time, later...)
   //
   PortableRef(Item * item, ItemPool * itemPool, RefPool * refPool, uint32 numItems = 1, bool doRefCounting = true)
   {
      NoteAllocated(this, "PortableRef");
      if (doRefCounting)
      {
         _rc = (refPool) ? refPool->GetObject() : newnothrow RefCountMem();
         if (_rc) 
         {
            _rc->Init(item, itemPool, refPool, numItems);
            _defaultPtr = NULL;
            _ownedItemCount = -1;  // _rc owns the pointer now
         }
         else
         {
            // can't allocate refCount!  Fall back to just holding it ourself (sole ownership)
            WARN_OUT_OF_MEMORY;
            _defaultPtr = item;
            _ownedItemCount = numItems;  // we own the pointer
         }
      }
      else
      {
         _rc = NULL;
         _defaultPtr = item;
         _ownedItemCount = -1;  // calling code still owns the pointer here
      }
   }

   // Create an additional reference to the object specified by (copyMe)
   PortableRef(PortableRef & copyMe) : _rc(NULL), _defaultPtr(NULL), _ownedItemCount(-1)
   {
      NoteAllocated(this, "PortableRef");
      *this = copyMe; 
   }

   // Creates a NULL reference (suitable for initializing with the = operator)
   PortableRef() : _rc(NULL), _defaultPtr(NULL), _ownedItemCount(-1) 
   {
      NoteAllocated(this, "PortableRef");
   }

   // Unreferences the held data
   ~PortableRef() 
   {
      Reset();
      NoteDeleted(this, "PortableRef");
   }

   // Unreferences the old held data, and references the new.
   PortableRef &operator=(const PortableRef &r) 
   {
      if (&r != this)
      {
         DeleteOwnedDefaultPointer();
         if (_rc) _rc->Unref();
         _rc = r._rc;
         if (_rc) _rc->Ref();
         _defaultPtr = (r._ownedItemCount >= 0) ? NULL : r._defaultPtr;  // if he is sole owner, we don't want it
      }
      return *this;
   }

   // Returns true iff both PortableRefs are referencing the same data
   int operator ==(const PortableRef &rhs) const {return _rc == rhs._rc;}
 
   // Returns the ref-counted data.  Please don't hold on to the
   // returned pointer as it is only guaranteed valid as long as this 
   // PortableRef object exists.
   Item * GetItemPointer() const {return _rc ? _rc->GetItem() : _defaultPtr;}

   // Returns the number of items in the array we are tracking (i.e.
   // one if we are tracking just a single object)
   uint32 GetItemCount() const {return _rc ? _rc->GetItemCount() : 0L;}

   // Convenience method:  Turns this object back into a NULL reference.
   // (same as *this = PortableRef();)
   void Reset() 
   {
      if (_rc) _rc->Unref();
      _rc = NULL;
      DeleteOwnedDefaultPointer();
   }
   
private:
   void DeleteOwnedDefaultPointer()
   {
      if (_ownedItemCount >= 0) 
      {
         if (_ownedItemCount == 1) delete _defaultPtr;
                               else delete [] _defaultPtr;
         _ownedItemCount = -1;
         _defaultPtr = NULL;
      }
   }

   //:This class is used internally by the PortableRef class.  An object of this class
   //:represents the reference count for a given object, and is shared by all the PortableRef
   //:classes referencing the tracked object.
   class RefCountMem
   { 
   public:
      RefCountMem() {NoteAllocated(this, "RefCountMem");} 

      void Init(Item * item, ItemPool * itemPool, RefPool * refPool, uint32 numItems)
      {
         _refCount  = 1;
         _item      = item;
         _itemPool  = itemPool;
         _refPool   = refPool;
         _numItems  = numItems;
      }

      ~RefCountMem() 
      {
         FreeObject();  // shouldn't be necessary but just in case
         NoteDeleted(this, "RefCountMem");
      }

      void FreeObject()
      {
         if (_item)
         {
            if (_itemPool) _itemPool->ReleaseObject(_item, _numItems);
            else 
            {
               if (_numItems == 1) delete _item;
                              else delete [] _item;
            }
            _item = NULL;
         }
      }
  
      void Ref()  
      {
         (void) ATOMIC_PREINCREMENT(_refCount); 
      }

      void Unref() 
      {
         if (ATOMIC_PREDECREMENT(_refCount) == 0) 
         {
            FreeObject();
            if (_refPool) _refPool->ReleaseObject(this);
                     else delete this;
         }
      }

      Item * GetItem() const {return _item;}

   private:
      Item * _item;
      int32 _refCount;
      RefPool * _refPool;
      ItemPool * _itemPool;
      uint32 _numItems;
   };

   RefCountMem * _rc;
   Item * _defaultPtr;  // used only when user is handling object deletion himself
   int _ownedItemCount;  // if >=0, we have sole ownership of (_defaultPtr), and _ownedItemCount is the array size
};

#endif
