/* This file is Copyright 2000 Level Control Systems.  See the included LICENSE.txt file for details. */

//! author="Jeremy Friesner"

#ifndef ABSTRACT_REFLECT_SESSION_H
#define ABSTRACT_REFLECT_SESSION_H

#include <stdio.h>
#include "support/PortableTypeDefs.h"
#include "dataio/TCPSocketDataIO.h"
#include "queue/PortableQueue.h"
#include "refcount/PortableRefCount.h"
#include "iogateway/AbstractMessageIOGateway.h"

class AbstractReflectSession;
class ReflectServer;

// Type declaration for a function that returns a new AbstractReflectSession object
typedef AbstractReflectSession* (*ReflectSessionFactoryFunc)();  

// Type declaration for a reference-count-pointer to an AbstractReflectSession
typedef PortableRef<AbstractReflectSession> AbstractReflectSessionRef;

//:This is the abstract base class that defines the server side of a client-server connection. 
// This class contains no message routing logic of its own, but defines
// the interface so that subclasses can do so.
class AbstractReflectSession
{
public:
   AbstractReflectSession();

   virtual ~AbstractReflectSession();

   // This method is called directly after the constructor.
   // Various methods that weren't callable from the constructor
   // may be called here.
   // Should return B_NO_ERROR if everything is okay; something
   // else if there is a problem and the session should be aborted.
   // Default implementation does nothing and returns B_NO_ERROR.
   virtual status_t AttachedToServer();

   // Returns the hostname of this session
   const char * GetHostName() const;

   // Returns the client-side port of this session
   int GetClientPort() const;

   // Returns a ID string to represent this session with.
   // The returned string is guaranteed to be unique across
   // all sessions in the server.
   virtual const char * GetSessionIDString() const;

   // Returns a reference to a PortableMessage that is shared by all sessions
   // This message can be used for whatever purpose the sessions care to; it
   // is not used by the server itself.
   PortableMessage & GetCentralState() const;

   // Returns an object that can be used to iterate over all the currently
   // connected sessions.
   const PortableQueue<AbstractReflectSessionRef> & GetSessionsList() const;

   // Returns a reference to the session with the given session ID, or B_ERROR.
   status_t GetSessionByID(const char * id, AbstractReflectSessionRef & returnRef) const;
   
   // Adds the given AbstractReflectSession to the session list.
   // This method is only needed in very specialized situations (e.g. 
   // if sessions are creating new sessions).
   // If (socket) is less than zero, no TCP connection will be used,
   // and the session will be a pure server-side entity.
   status_t AddNewSession(AbstractReflectSessionRef session, int socket);

   // Marks this session to be terminated, as if the TCP connection had been broken.
   void EndSession();

   // Causes this session to be terminated (similar to EndSession(), 
   // and the session specified in (newSessionRef) to take its 
   // place using the same socket connection & message IO gateway.  
   // Returns B_NO_ERROR on success, B_ERROR if the new session refused
   // to be attached.
   status_t ReplaceSession(AbstractReflectSessionRef newSession);

   // Called when a new message is received from our IO gateway.
   virtual void MessageReceivedFromGateway(PortableMessageRef msg) = 0;

   // Called when a message is sent to us by another session (possibly this one).  
   // (from) is the session who sent the message to us.
   // (msg) is a reference to the message that was sent.
   // (userData) is additional data that is defined by the sending subclass.
   // (For StorageReflectSessions, this value, if non-NULL, is a pointer to the 
   // DataNode in this Session's node subtree that was matched by the paths in (msg))
   virtual void MessageReceivedFromNeighbor(AbstractReflectSession & from, PortableMessageRef msg, void * userData) = 0;

   // Called when the TCP connection to our client is broken.
   // If this method returns true, then this session will be removed and
   // deleted.  If it returns false, then this session will continue, even
   // though the client is not available.  Default implementation returns true.
   virtual bool ClientConnectionClosed();
  
protected:
   // Adds (ref) to our gateway's outgoing messsage queue.
   // (ref) will be sent back to our client when time permits.
   status_t AddOutgoingMessage(PortableMessageRef ref);

   // Gateway factory method.  Should return a new AbstractMessageIOGateway
   // to use to communicate over the given socket.  The returned gateway
   // object is responsible for closing the socket when it is done with it.
   // The default implementation returns a regular PortableMessageIOGatway with
   // non-blocking TCPSocketDataIO.
   virtual AbstractMessageIOGateway * CreateGateway(int socket);

   // Should return a pretty human readable string identifying this class.
   // (typeid.name() isn't pretty enough)
   virtual const char * GetTypeName() const = 0;

   // Returns true iff we are attached to the server
   bool IsAttachedToServer() const {return (_owner != NULL);}

   // Returns the held message IO gateway object, or NULL if there is none.
   // The returned object remains the property of this session.
   AbstractMessageIOGateway * GetGateway() const {return _gateway;}

   // Returns the socket descriptor our gateway is using, or -1.
   int GetSocket() const {return _socket;}

   // Replaces the held message IO gateway object with (gw).
   // If (deleteOldGateway) is true, then the old gateway object 
   // (if any) will be deleted.   (gw) becomes the 
   // property of this session object.  (gw) must have been allocated 
   // with new.
   // (newSocketFD) should be the TCP socket descriptor that the new gateway is 
   // using, or -1 if (gw) is NULL.
   void SetGateway(AbstractMessageIOGateway * gw, int newSocketFD, bool deleteOldGateway = true);

private:
   friend class ReflectServer;

   char _idString[20];

   // The following variables are set directly by the ReflectServer.
   ReflectServer * _owner;
   int _socket;  
   int _port;
   PortableString _hostName;
   AbstractMessageIOGateway * _gateway; /* we own this object */
};

#endif
