/* This file is Copyright 2000 Level Control Systems.  See the included LICENSE.txt file for details. */

//! author="Jeremy Friesner"

#ifndef PORTABLE_REFLECT_SERVER_H
#define PORTABLE_REFLECT_SERVER_H

#include "reflector/AbstractReflectSession.h"

// Add a callback function that should be called when we are out of memory.
// Functions added to this list can try to free up memory, etc.
typedef void (*OutOfMemoryHandlerFunc)();
extern void AddOutOfMemoryHandler(OutOfMemoryHandlerFunc f);

//: This class represents a reflection server:  It runs on a centrally located machine,
//: and many clients may connect to it simultaneously.  This server can then redirect messages
//: uploaded by any client to other clients in a somewhat efficient manner.
class ReflectServer
{
public: 
   // (maxBytes) specifies the maximum number of bytes of memory the server will allow it
   // self to have allocated at once.  Defaults to somewhat over 4 gigabytes.
   ReflectServer(uint32 maxBytes = ((uint32)-1L));
   virtual ~ReflectServer();

   //:The main loop for the message reflection server.
   // Returns B_ERROR if there was a problem setting up.  This method will not return until the server stops running (usually due to an error)
   //  (acceptPort) is the port the server will listen on.
   //  (createSessionFunc) should be a function of the form:
   //     AbstractReflectSession * MyFunc();
   //  that returns a new AbstractReflectSession object to be used for communicating with 
   //  the given socket.  The object this function returns determines the semantics of the server.
   //  Most non-abstract ReflectSession classes come with such a function.
   virtual status_t ServerProcessLoop(ReflectSessionFactoryFunc createSessionFunc, short acceptPort = 2960);

   //: Called after the server is set up, but just before accepting any connections.
   // Should return B_NO_ERROR if it's okay to continue, or B_ERROR to abort.
   // Default implementation returns B_NO_ERROR.
   //
   virtual status_t ReadyToRun();

   // Should be called just before the ReflectServer is to be destroyed.
   // this in a good place to put any cleanup code.  Be sure
   // to call Cleanup() of your parent class as well!
   // (We can't just do this in the destructor, as some cleanup
   // relies on the subclass still being functional, which it isn't
   // when our destructor gets called!)
   virtual void Cleanup();

protected:
   // Adds a new session that uses the given socket for I/O.
   // If (socket) is -1, no TCP connection will be used. 
   virtual status_t AddNewSession(AbstractReflectSessionRef ref, int socket);

   // This version of AddNewSession (which is called by the previous 
   // version) assumes that the gateway, hostname, port, etc of the
   // new session have already been set up.
   virtual status_t AddNewSession(AbstractReflectSessionRef ref);

   // Returns a human-readable string that describes the type of server that
   // is running.  Default implementation returns "MUSCLE".
   virtual const char * GetServerName() const;

   // May be overridden to specify when Pulse() should next be called.
   // (now) is the current time of day.
   // (returnTime) should be modified to reflect the time at which Pulse() should
   // next be called.  If it is set to (0,0), then no Pulse() call will be scheduled.
   // Default version returns the next schedule time for any attached i/o gateway
   // that wants Pulse() events.  If you override this method, please call up to
   // this method implementation, and only change the return value yourself if your
   // wakeup time is before the (returnTime) you get.
   virtual void GetPulseTime(const struct timeval & now, struct timeval & returnTime);

   // If GetPulseTime() was overridden, then this callback will be called at
   // (or shortly after) the time that GetPulseRate() specified.
   // (now) is the current time of day.  Default implementation calls Pulse()
   // on any i/o gateways that scheduled Pulse() events.
   virtual void Pulse(const struct timeval & now);

   // Returns our contained list of active sessions.
   PortableQueue<AbstractReflectSessionRef> & GetSessionsList();

   // Called by a session to get itself removed & destroyed
   void EndSession(AbstractReflectSession * which);

   // Called by a session to get itself replaced (the
   // new session will continue using the same message io streams
   // as the old one)
   // Returns B_NO_ERROR on success, B_ERROR if the new session
   // returns an error in its AttachedToServer() method.  If 
   // B_ERROR is returned, then this call is guaranteed not to
   // have had any effect on the old session.
   status_t ReplaceSession(AbstractReflectSessionRef newSession, AbstractReflectSession * replaceThisOne);

   // Accessor to our central-state repository message
   PortableMessage & GetCentralState() {return _centralState;}

private:
   friend class AbstractReflectSession;

   PortableMessage _centralState;
   PortableQueue<AbstractReflectSessionRef> _sessions;
   PortableHashtable<const char *, AbstractReflectSessionRef> _idLookup;
   PortableQueue<AbstractReflectSessionRef> _lameDuckSessions;

   struct timeval _nextPulseAt;
};

#endif
