/* This file is Copyright 2000 Level Control Systems.  See the included LICENSE.txt file for details. */

#include <stdio.h>
#include <stdarg.h>
#include <time.h>
#include "support/PortableMacros.h"
#include "syslog/SysLog.h"
#include "queue/PortableQueue.h"

#ifdef WATCH_MUSCLE_MEM_USAGE

#include "hashtable/PortableHashtable.h"
#include "string/PortableString.h"
#include "netutil/NetworkUtilityFunctions.h"

static PortableHashtable<PortableString, uint64> * _allocTable = NULL;

static struct timeval _nextPrintAt = {0, 0};
static bool _isNotingAllocation = false;  // avoids infinite regress when NoteAllocated uses hashtable, string, etc.
static uint32 _logID = 0;

void NoteAllocated(void * This, const char * className)
{
   if (_isNotingAllocation == false)
   {
      bool printLog = false;

      _isNotingAllocation = true;
      {
         const PortableString cs(className); 

         uint64 prevVal = 0LL;
         if (_allocTable == NULL) _allocTable = newnothrow PortableHashtable<PortableString, uint64>(PortableStringHashFunc);
         if (_allocTable)
         {
            (void) _allocTable->Get(cs, prevVal);
            _allocTable->Put(cs, prevVal+1);
         }
         else WARN_OUT_OF_MEMORY;

         LogTime(LOG_TRACE, "@Alloced obj=%p (%s) (%i)\n", This, className, prevVal+1);

         const struct timeval interval = {WATCH_MUSCLE_MEM_USAGE, 0};
         printLog = OnceEvery(interval, _nextPrintAt);
      }
      _isNotingAllocation = false;
      if (printLog) LogCurrentAllocations();
   }
}

void NoteDeleted(void * This, const char * className)
{
   if (_isNotingAllocation == false)
   {
      _isNotingAllocation = true;
      {
         uint64 prevVal;
         PortableString cs(className);
         if ((_allocTable)&&(_allocTable->Get(cs, prevVal)&&(prevVal > 0)))
         {
            if (prevVal > 1) _allocTable->Put(cs, prevVal-1);
                        else _allocTable->Remove(cs);

            LogTime(LOG_TRACE, "@Deleted obj=%p (%s) (%i)\n", This, className, prevVal-1);
            if (_allocTable->Size() == 0)
            {
               delete _allocTable;
               _allocTable = NULL;
            }
         }
         else
         {
            PortableString errMsg("NoteDeleted:  Objects of class ");
            errMsg += className;
            errMsg += " deleted too many times!?";
            MASSERT(false, errMsg.Cstr());
         }
      }
      _isNotingAllocation = false;
   }
}

void LogCurrentAllocations()
{
   if (_isNotingAllocation == false)
   {
      _logID++;
      _isNotingAllocation = true;
      LogTime(LOG_INFO, "=====CURRENT OBJECT STATISTICS (%i)=======\n", _allocTable?_allocTable->Size():0);
      if (_allocTable)
      {
         PortableHashtableIterator<PortableString, uint64> iter = _allocTable->GetIterator();
         PortableString * nextKey;
         while((nextKey = iter.GetNextKey()) != NULL) LogTime(LOG_INFO, "  %s: %Lu\n", nextKey->Cstr(), *iter.GetNextValue());
      }
      _isNotingAllocation = false;
   }
}

#endif  // WATCH_MUSCLE_MEM_USAGE

static const char * const _logLevelNames[] = {
   "None",
   "Critical Errors Only",
   "Errors Only",
   "Warnings and Errors Only",
   "Informational",
   "Debug",
   "Trace"
};

class DefaultLogger : public LogCallback
{
public:
   DefaultLogger()
      : _fileLogLevel(LOG_NONE), _consoleLogLevel(LOG_INFO), _logFile(NULL)
   {
       // empty
   }

   virtual void Log(time_t when, int ll, const char * fmt, va_list argList)
   {
      if ((ll <= _fileLogLevel)&&(_logFile == NULL))
      {                                          
         struct tm * now = localtime(&when);
         char tb[100];                           
         sprintf(tb, "%02i%02i%02i%02i.log", now->tm_mon+1, now->tm_mday, now->tm_hour, now->tm_min); 
         _logFile = fopen(tb, "w");              
         if (_logFile) LogTime(LOG_INFO, "Created log file %s\n", tb);
      }                                          
      if (ll <= _consoleLogLevel) vprintf(fmt, argList);                  
      if ((ll <= _fileLogLevel)&&(_logFile))
      {                                         
         vfprintf(_logFile, fmt, argList);       
         fflush(_logFile);                       
      }                                          
   }

   int _fileLogLevel;
   int _consoleLogLevel;
   FILE * _logFile;
};

static PortableQueue<LogCallback *> _logCallbacks;
static DefaultLogger _dl;

const char * GetLogLevelName(int ll)
{
   return ((ll>=0)&&(ll<(int) ARRAYITEMS(_logLevelNames))) ? _logLevelNames[ll] : "???";
}

int GetFileLogLevel()
{
   return _dl._fileLogLevel;
}

int GetConsoleLogLevel()
{
   return _dl._consoleLogLevel;
}

int GetMaxLogLevel()
{
   return (_dl._consoleLogLevel > _dl._fileLogLevel) ? _dl._consoleLogLevel : _dl._fileLogLevel;
}

void SetFileLogLevel(int loglevel)
{
   _dl._fileLogLevel = loglevel;
   LogTime(LOG_INFO, "File logging level set to: %s\n", GetLogLevelName(_dl._fileLogLevel));
}

void SetConsoleLogLevel(int loglevel)
{
   _dl._consoleLogLevel = loglevel;
   LogTime(LOG_INFO, "Console logging level set to: %s\n", GetLogLevelName(_dl._consoleLogLevel));
}

#define DO_LOGGING                                    \
{                                                     \
   time_t when = time(NULL);                          \
   va_list argList;                                   \
   va_start(argList, fmt);                            \
   _dl.Log(when, ll, fmt, argList);                   \
   va_end(argList);                                   \
   int numLogCallbacks = _logCallbacks.CountItems();  \
   for (int i=0; i<numLogCallbacks; i++)              \
   {                                                  \
      LogCallback * next = _logCallbacks[i];          \
      va_start(argList, fmt);                         \
      next->Log(when, ll, fmt, argList);              \
      va_end(argList);                                \
   }                                                  \
}

void LogTime(int ll, const char * fmt, ...)
{
   time_t n = time(NULL);
   struct tm * now = localtime(&n);

   // Record time stamp to our file/stdout logs only
   char buf[128];
   sprintf(buf, "[%c %02i/%02i %02i:%02i:%02i] ", GetLogLevelName(ll)[0], now->tm_mon+1, now->tm_mday, now->tm_hour, now->tm_min, now->tm_sec);
   va_list dummyList;
   va_start(dummyList, fmt);  // not used
   _dl.Log(n, ll, buf, dummyList);
   va_end(dummyList);
   
   // Log message to file/stdio and callbacks
   DO_LOGGING;
}

void Log(int ll, const char * fmt, ...)
{
   DO_LOGGING;
}

void AddLogCallback(LogCallback * cb)
{
   _logCallbacks.AddTail(cb);
}

void RemoveLogCallback(LogCallback * cb, bool deleteToo)
{
   for (int i=_logCallbacks.CountItems()-1; i>=0; i--)
   {
      LogCallback * next = _logCallbacks[i];
      if ((next == cb)||(cb == NULL))
      {
         _logCallbacks.RemoveItemAt(i);
         if (deleteToo) delete next;
         if (cb) break;
      }
   }
}
