/* This file is Copyright 2000 Level Control Systems.  See the included LICENSE.TXT file for details. */

//! author="Jeremy Friesner"

#ifndef MUSCLE_SYS_LOG_H
#define MUSCLE_SYS_LOG_H

#include <time.h>
#include <stdarg.h>

//:log level constants to use with SetLogLevel(), GetLogLevel()
enum
{
   LOG_NONE = 0,         // nothing ever gets logged at this level (default)
   LOG_CRITICALERROR,    // only things that should never ever happen
   LOG_ERROR,            // things that shouldn't usually happen
   LOG_WARNING,          // things that are suspicious
   LOG_INFO,             // things that the user might like to know
   LOG_DEBUG,            // things the programmer is debugging with
   LOG_TRACE,            // exhaustively detailed output
   NUM_LOGLEVELS
}; 

//:Returns the current log level for logging to a file.
int GetFileLogLevel();

//:Returns the current log level for logging to stdout.
int GetConsoleLogLevel();

//:Returns the max of GetFileLogLevel() and GetConsoleLogLevel()
int GetMaxLogLevel();

//:Sets the log filter level for logging to a file.  
//Any calls to Log*() that specify a log level greater than (loglevel)
//will be suppressed.  Default level is LOG_NONE (i.e. no file logging is done)
void SetFileLogLevel(int loglevel);

//:Sets the log filter level for logging to stdout.
//Any calls to Log*() that specify a log level greater than (loglevel)
//will be suppressed.  Default level is LOG_INFO.
void SetConsoleLogLevel(int loglevel);

//:Same semantics as printf, only outputs to the log file/console instead
void Log(int minLogLevel, const char * fmt, ...);

//:Formatted.  Automagically prepends timestand, appends carriage return
void LogTime(int minLogLevel, const char * fmt, ...);

//:Returns a human-readable string for the given log level.
const char * GetLogLevelName(int logLevel);

//:Callback object that can be added with AddLogCallback()
class LogCallback
{
public:
   virtual ~LogCallback() {/* empty */}
   virtual void Log(time_t when, int logLevel, const char * format, va_list argList) = 0;
};

//:Add a custom log callback object.
void AddLogCallback(LogCallback * cb);

//:Removes the given callback from our list.  
//If (cb) is NULL, removes ALL LogCallbacks.
//If (deleteToo) is true, any removed callback(s) will be deleted as well.
void RemoveLogCallback(LogCallback * cb, bool deleteToo);

#ifdef WATCH_MUSCLE_MEM_USAGE
void NoteAllocated(void * This, const char * className);
void NoteDeleted(void * This, const char * className);
void LogCurrentAllocations();
#else
# define NoteAllocated(t,x)
# define NoteDeleted(t,x)
# define LogCurrentAllocations()
#endif

#endif
